/** 3DGPL *************************************************\
 * (NeXTStep, 8bit deep modes, 32bit mashine)             *
 * Header for hardware specific stuff.                    *
 *                                                        *
 * Defines:                                               *
 *  HW_set_int               Should be a macro;           *
 *  HW_set_16_bit            And this one.                *
 *                                                        *
 *  HW_init_screen           Opening output surface;      *
 *  HW_blit                  Colourmap onto the screen;   *
 *  HW_close_screen          Closing output;              *
 *                                                        *
 *  HW_init_event_loop       Runiing for events;          *
 *  HW_error                 Quiting with a message;      *
 *  HW_close_event_loop      Quiting running.             *
 *                                                        *
 * Internal:                                              *
 *  HWI_colour_norm          Norm of the rgb vector.      *
 *                                                        *
 * (c) 1995-98 Sergei Savchenko, (savs@cs.mcgill.ca)      *
\**********************************************************/

#include "../hardware/hardware.h"           /* harware dependent stuff */
#include "../colour/colour.h"               /* CL_ stuff */
#include "../graphics/graphics.h"           /* G_ stuff */
#import  <appkit/appkit.h>                  /* id type */
#import  <dpsclient/wraps.h>                /* DPS functions */
#import  <dpsclient/event.h>                /* Event mascks */
#include <stdio.h>                          /* vsptrintf */
#include <stdarg.h>                         /* var args */
#include <stdlib.h>                         /* exit */
#include <math.h>                           /* sqrt */

int HW_running=0;                           /* event loop on? */
id HW_window;                               /* application's main window */
id HW_view;                                 /* the "view" of the window */
NXRect HW_rect;                             /* screen boundaries */
DPSTimedEntry HW_entry;                     /* initialized timed entry */

void (*HW_application_main)(void);          /* pointers to user functions */
void (*HW_application_key_handler)(int key_code);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Norm of the RGB colour.                               *
 *                                                       *
 * RETURNS: norm of the RGB vector.                      *
 * --------                                              *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * */

float HWI_colour_norm(struct CL_colour *colour)
{
 return(sqrt(((float)colour->cl_red)*colour->cl_red+
             ((float)colour->cl_green)*colour->cl_green+
             ((float)colour->cl_blue)*colour->cl_blue
            )
       );
}

/**********************************************************\
 * Implementations for fast memory operation.             *
\**********************************************************/

void HW_set_int(int *d,long l,int v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Implementations for another fast memory operation.     *
\**********************************************************/

void HW_set_16_bit(HW_16_bit *d,long l,HW_16_bit v)
{
 long i;

 for(i=0;i<l;i++) *d++=v;
}

/**********************************************************\
 * Creating a window.                                     *
\**********************************************************/

#define HW_MAX_COLOUR_VALUE   255           /* 8 bit, 2 power 8 */
#define HW_TOTAL_COLOUR_CELLS 256           /* actually can do more then that */

void HW_init_screen(char *display_name,
                    char *screen_name
		   )
{
 int i;
 float remap;                               /* parameter for rescaling */

 [Application new];

 NXSetRect(&HW_rect,360.0,300.0,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE);
 HW_window = [[Window alloc] initContent:&HW_rect
				   style:NX_TITLEDSTYLE
				 backing:NX_BUFFERED
			      buttonMask:NX_MINIATURIZEBUTTONMASK
				   defer:NO];

 [HW_window setTitle:screen_name];
 [HW_window display];
 [HW_window orderFront:nil];

 HW_view=[HW_window contentView];            /* window's main view */
 NXSetRect(&HW_rect,0.0,0.0,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE);
 [HW_view allocateGState];

 remap=(float)HW_TOTAL_COLOUR_CELLS/         /* remapping into flat b/w space */
       sqrt(3*HW_MAX_COLOUR_VALUE*(float)HW_MAX_COLOUR_VALUE);

 for(i=0;i<(CL_COLOUR_LEVELS*CL_LIGHT_LEVELS);i++)
  CL_colours->cl_intensity_table[i]=
   HWI_colour_norm(CL_colours->cl_colours+CL_colours->cl_intensity_table[i])*remap;
}

/**********************************************************\
 * Blitting a bitmap into the allocated window.           *
\**********************************************************/

void HW_blit(void)
{
 [HW_view lockFocus];                       /* lock output on the view */

 NXImageBitmap(&HW_rect,HW_SCREEN_X_SIZE,HW_SCREEN_Y_SIZE,8,1,NX_MESHED,
	       NX_MONOTONICMASK,G_c_buffer,NULL,NULL,NULL,NULL
	      );

 [HW_window flushWindow];                   /* image appears on screen */
 [HW_view unlockFocus];                     /* unlock output */
}

/**********************************************************\
 * Deallocating a window.                                 *
\**********************************************************/

void HW_close_screen(void)
{
}

/**********************************************************\
 * The main event loop.                                   *
\**********************************************************/

void HW_event_loop()
{
 NXEvent *anEvent;

 HW_application_main();

 while((anEvent=[NXApp getNextEvent:NX_KEYDOWNMASK
			 waitFor:0.0
		       threshold:NX_BASETHRESHOLD])!=NULL)
  HW_application_key_handler(anEvent->data.key.charCode);
}

/**********************************************************\
 * Running the event loop.                                *
\**********************************************************/

void HW_init_event_loop(void (*application_main)(void),
			void (*application_key_handler)(int key_code)
		       )
{
 HW_running=1;
 [HW_window makeKeyWindow];                 /* don't wanna write any else */

 HW_application_main=application_main;
 HW_application_key_handler=application_key_handler;
 HW_entry=DPSAddTimedEntry(0.0,&HW_event_loop,NULL,NX_BASETHRESHOLD);

 [NXApp run];
 [NXApp free];                              /* run and be free */
 HW_running=0;
}

/**********************************************************\
 * Quiting with a message.                                *
\**********************************************************/

#define HW_MAX_ERROR_MESSAGE 256

void HW_error(char *s,...)
{
 char str[HW_MAX_ERROR_MESSAGE];
 va_list lst;

 va_start(lst,s);
 vsprintf(str,s,lst);                       /* forming the message */
 va_end(lst);

 if(HW_running)                             /* safe to show alert panel */
 {
  NXRunAlertPanel("3Dgpl",str,NULL,NULL,NULL);
  HW_close_event_loop();
 }
 else
 {
  fprintf(stderr,str);                      /* probably not safe */
  exit(0);
 }
}

/**********************************************************\
 * Quiting the event loop.                                *
\**********************************************************/

void HW_close_event_loop(void)
{
 DPSRemoveTimedEntry(HW_entry);
 [NXApp terminate:0];   
}

/**********************************************************/
